use std::{collections::HashMap, sync::LazyLock};

use bollard::{
    Docker,
    image::CreateImageOptions,
    query_parameters::{
        CreateContainerOptionsBuilder, CreateImageOptionsBuilder, ListContainersOptionsBuilder,
        StartContainerOptions, StartContainerOptionsBuilder,
    },
    secret::{ContainerCreateBody, HostConfig, PortBinding},
};
use futures_util::TryStreamExt;

use crate::{models::port::Port, state::ClientState};

#[derive(Clone)]
pub struct ContainerRuntime {
    client: Docker,
}

impl ContainerRuntime {
    #[tracing::instrument(skip(self), level = "trace")]
    pub async fn ensure_running(
        &self,
        name: &str,
        image: &str,
        ports: Vec<Port>,
    ) -> anyhow::Result<()> {
        tracing::debug!("ensuring that image is running");

        let containers = self
            .client
            .list_containers(Some(
                ListContainersOptionsBuilder::default()
                    .all(true)
                    .filters(&HashMap::from([(
                        "name".to_string(),
                        vec![name.to_string()],
                    )]))
                    .build(),
            ))
            .await?;
        if !containers.is_empty() {
            // Reconcile difference
            return Ok(());
        }

        let _ = self
            .client
            .create_image(
                Some(CreateImageOptionsBuilder::new().from_image(image).build()),
                None,
                None,
            )
            .try_collect::<Vec<_>>()
            .await?;

        let ports: HashMap<_, _> = ports
            .iter()
            .map(|p| {
                (
                    format!("{}/tcp", p.container_port),
                    Some(vec![PortBinding {
                        host_ip: Some("0.0.0.0".into()),
                        host_port: Some(p.host_port.to_string()),
                    }]),
                )
            })
            .collect();
        self.client
            .create_container(
                Some(CreateContainerOptionsBuilder::new().name(name).build()),
                ContainerCreateBody {
                    image: Some(image.into()),
                    host_config: Some(HostConfig {
                        port_bindings: Some(ports),
                        ..Default::default()
                    }),
                    ..Default::default()
                },
            )
            .await?;

        self.client
            .start_container(name, Some(StartContainerOptionsBuilder::default().build()))
            .await?;

        Ok(())
    }
}

pub trait ContainerRuntimeState {
    fn container_runtime(&self) -> ContainerRuntime;
}

impl ContainerRuntimeState for ClientState {
    fn container_runtime(&self) -> ContainerRuntime {
        static CLIENT: LazyLock<bollard::Docker> = LazyLock::new(|| {
            Docker::connect_with_defaults().expect("to be able to connect to a docker daemon")
        });

        ContainerRuntime {
            client: CLIENT.clone(),
        }
    }
}
